LIMITS = {'BNB/BTC': {'amount': {'max': 90000000.0, 'min': 0.01},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'BNB/ETH': {'amount': {'max': 90000000.0, 'min': 0.01},
                      'cost': {'max': None, 'min': 0.01},
                      'price': {'max': None, 'min': None}},
          'BNB/USDT': {'amount': {'max': 10000000.0, 'min': 0.01},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'BTC/USDT': {'amount': {'max': 10000000.0, 'min': 1e-06},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'ETH/BTC': {'amount': {'max': 100000.0, 'min': 0.001},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'ETH/USDT': {'amount': {'max': 10000000.0, 'min': 1e-05},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'XRP/BNB': {'amount': {'max': 90000000.0, 'min': 0.1},
                      'cost': {'max': None, 'min': 1.0},
                      'price': {'max': None, 'min': None}},
          'XRP/BTC': {'amount': {'max': 90000000.0, 'min': 1.0},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'XRP/ETH': {'amount': {'max': 90000000.0, 'min': 1.0},
                      'cost': {'max': None, 'min': 0.01},
                      'price': {'max': None, 'min': None}},
          'XRP/USDT': {'amount': {'max': 90000000.0, 'min': 0.1},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'XLM/USDT': {'amount': {'max': 90000000.0, 'min': 0.1},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'XLM/XRP': {'amount': {'max': 90000000.0, 'min': 0.1},
                      'cost': {'max': None, 'min': 1.0},
                      'price': {'max': None, 'min': None}}}


class DummyExchange():

    def __init__(self, currencies, balances, rates=None, fee=0.001):
        self.name = 'DummyExchange'
        self._currencies = currencies
        self._balances = balances
        self._fee = fee
        self._rates = {}
        for cur in rates or {}:
            self._rates[cur] = {'mid': rates[cur],
                                'high': rates[cur]*1.001,
                                'low': rates[cur]*0.999,
                                }
        
    @property
    def balances(self):
        return self._balances

    @property
    def pairs(self):
        _pairs = []
        for i in self._currencies:
            for j in self._currencies:
                pair = "{}/{}".format(i, j)
                _pairs.append(pair)
        return _pairs

    @property
    def rates(self):
        if self._rates:
            return self._rates

        _rates = {}
        for pair in self.pairs:
            _rates[pair] = {'mid': 1.0,
                            'low': 0.99,
                            'high': 1.01, }
        return _rates

    @property
    def limits(self):
        return LIMITS

    @property
    def fee(self):
        return self._fee

    def preprocess_order(self, order):
        try:
            limits = self.limits[order.pair]
        except KeyError:
            return None
        if order.amount < limits['amount']['min'] \
           or order.amount * order.price < limits['cost']['min']:
            return None

        base, quote = order.pair.split('/')
        if order.direction.upper() == 'BUY':
            if order.amount * order.price > self._balances[quote]:
                return None

        if order.direction.upper() == 'SELL':
            if order.amount > self._balances[base]:
                return None

        order.type_ = 'LIMIT'
        return order

    def execute_order(self, order):
        base, quote = order.pair.split('/')
        if order.direction.upper() == 'BUY':
            if order.amount * order.price > self._balances[quote]:
                raise ValueError("Can't overdraw")
            self._balances[base] += order.amount
            self._balances[base] -= order.amount * self.fee
            self._balances[quote] -= order.amount * order.price

        if order.direction.upper() == 'SELL':
            if order.amount > self._balances[base]:
                raise ValueError("Can't overdraw")
            self._balances[base] -= order.amount
            self._balances[quote] += order.amount * order.price
            self._balances[quote] -= order.amount * order.price * self.fee

        return {'symbol': order.pair,
                'side': order.direction.upper(),
                'amount': order.amount,
                'price': order.price}



import logging

from crypto_reflex.order import Order

logger = logging.getLogger(__name__)


class Executor():

    def __init__(self, portfolio, exchange, balancer):
        self.portfolio = portfolio
        self.exchange = exchange
        self.balancer = balancer

    def run(self, force=False, trade=False, max_orders=5, mode='mid'):

        balances = self.portfolio.balances

        res = {'orders': [],
               'success': [],
               'errors': [],
               'balances': balances,
               'total_fee': 0.0,
               'initial_portfolio': self.portfolio,
               'proposed_portfolio': None, }

        if self.portfolio.needs_balancing or force:
            orders = self.balancer.balance(self.portfolio,
                                           self.exchange,
                                           max_orders,
                                           mode)

            if orders['proposed_portfolio']:
                res['proposed_portfolio'] = orders['proposed_portfolio']
                res['total_fee'] = orders['total_fee']
                res['orders'] = orders['orders']

                if trade:
                    for order in orders['orders']:
                        try:
                            r = self.exchange.execute_order(order)
                            res['success'].append(Order(r['symbol'],
                                                        r['side'].upper(),
                                                        r['amount'],
                                                        r['price']))
                        except Exception as e:
                            res['errors'].append(order)
                            logger.error("Could not place order: {} {}"
                                         .format(order, e))

        return res



import unittest
from pstats import Stats
import cProfile

from crypto_reflex.simple_balancer import SimpleBalancer
from crypto_reflex.portfolio import Portfolio
from crypto_reflex.dummy_exchange import DummyExchange
from crypto_reflex.executor import Executor
from crypto_reflex.order import Order

import sys
sys.path.append('..')      # XXX Probably needed to import your code


class test_Order(unittest.TestCase):

    def test_createOrder(self):
        order = Order('XRP/USDT', 'BUY', 10, 0.32)
        self.assertEqual(order.pair, 'XRP/USDT')
        self.assertEqual(order.direction, 'BUY')
        self.assertEqual(order.amount, 10)
        self.assertEqual(order.price, 0.32)

        order = Order('XRP/BTC', 'SELL', 20, 0.0001)
        self.assertEqual(order.pair, 'XRP/BTC')
        self.assertEqual(order.direction, 'SELL')
        self.assertEqual(order.amount, 20)
        self.assertEqual(order.price, 0.0001)

    def test_createOrderBadDirection(self):
        with self.assertRaises(ValueError):
            Order('XRP/USDT', 'FOO', 10, 0.001)

    def test_compareOrders(self):
        a = Order('XRP/USDT', 'BUY', 10, 0.32)
        b = Order('XRP/USDT', 'BUY', 10, 0.32)
        self.assertEqual(a, b)

        c = Order('XRP/BTC', 'SELL', 20, 0.0001)
        self.assertNotEqual(a, c)

        self.assertLess(c, a)
        self.assertGreater(a, c)

    def test_ReprStrOrder(self):
        a = Order('XRP/USDT', 'BUY', 10, 0.32)
        self.assertEqual(str(a), 'BUY 10.0 XRP/USDT @ 0.32')
        self.assertEqual(repr(a), "Order('XRP/USDT', 'BUY', 10.0, 0.32)")

    def test_HashOrder(self):
        a = Order('XRP/USDT', 'BUY', 10, 0.32)
        b = Order('XRP/USDT', 'BUY', 10, 0.32)
        c = Order('XLM/USDT', 'BUY', 10, 0.32)
        d = Order('XLM/USDT', 'SELL', 10, 0.32)
        e = Order('XLM/USDT', 'SELL', 100, 0.32)
        f = Order('XLM/USDT', 'SELL', 100, 0.1)
        self.assertEqual(hash(a), hash(b))
        self.assertNotEqual(hash(a), hash(f))
        self.assertNotEqual(hash(b), hash(c))
        self.assertNotEqual(hash(c), hash(d))
        self.assertNotEqual(hash(d), hash(e))
        self.assertNotEqual(hash(e), hash(f))


class test_Portfolio(unittest.TestCase):
    targets = {'XRP': 45,
               'XLM': 45,
               'USDT': 10, }

    targets2 = {'XRP': 40,
                'XLM': 40,
                'USDT': 20, }

    balances = {'XRP': 450,
                'XLM': 450,
                'USDT': 100, }

    zero_balances = {'XRP': 0,
                     'XLM': 0,
                     'USDT': 0, }

    def test_create_portfolio_defaults(self):
        exchange = DummyExchange(self.targets.keys(), self.targets)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.threshold, 1.0)
        self.assertEqual(portfolio.quote_currency, 'USDT')
        self.assertEqual(portfolio.exchange, exchange)
        self.assertEqual(portfolio.targets, self.targets)

    def test_create_portfolio_custom(self):
        exchange = DummyExchange(self.targets.keys(), self.targets)
        portfolio = Portfolio.make_portfolio(self.targets,
                                             exchange, 2.0, 'BTC')

        self.assertEqual(portfolio.threshold, 2.0)
        self.assertEqual(portfolio.quote_currency, 'BTC')
        self.assertEqual(portfolio.exchange, exchange)
        self.assertEqual(portfolio.targets, self.targets)

    def test_create_portfolio_balances_quote(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.balances_quote, self.balances)

    def test_create_portfolio_valuation_quote(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.valuation_quote, 1000)

    def test_create_portfolio_balances_pct(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.balances_pct, self.targets)
        self.assertNotEqual(portfolio.balances_pct, self.targets2)

    def test_create_portfolio_balances_pct_zero(self):
        exchange = DummyExchange(self.targets.keys(), self.zero_balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.balances_pct, self.zero_balances)
        self.assertNotEqual(portfolio.balances_pct, self.targets)

    def test_create_portfolio_metric1(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.balance_rms_error, 0)

    def test_create_portfolio_metric2(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets2, exchange)

        self.assertAlmostEqual(portfolio.balance_rms_error, 7.071067, 5)

    def test_create_portfolio_metric_zero(self):
        exchange = DummyExchange(self.targets.keys(), self.zero_balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertEqual(portfolio.balance_rms_error, 0)

    def test_create_portfolio_differences_quote1(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        expected = {'XRP': 0,
                    'XLM': 0,
                    'USDT': 0, }

        self.assertEqual(portfolio.differences_quote, expected)

    def test_create_portfolio_differences_quote2(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets2, exchange)

        expected = {'XRP': -50,
                    'XLM': -50,
                    'USDT': 100, }

        self.assertEqual(portfolio.differences_quote, expected)

    def test_create_portfolio_needs_balancing1(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets, exchange)

        self.assertFalse(portfolio.needs_balancing)

    def test_create_portfolio_needs_balancing2(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets2, exchange)

        self.assertTrue(portfolio.needs_balancing)

    def test_create_portfolio_needs_balancing3(self):
        exchange = DummyExchange(self.targets.keys(), self.balances)
        portfolio = Portfolio.make_portfolio(self.targets2,
                                             exchange, threshold=20)

        self.assertFalse(portfolio.needs_balancing)

    def test_base_differences_start_xrp_xlm_usdt_rates(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 100,
                   'XLM': 100,
                   'USDT': 50, }
        rates = {'XRP/USDT': 0.1,
                 'XLM/USDT': 0.4, }

        exchange = DummyExchange(targets.keys(), current, rates)
        portfolio = Portfolio.make_portfolio(targets, exchange)

        expected = {'XRP': 40,
                    'XLM': 0,
                    'USDT': -40, }
        self.assertEqual(portfolio.differences_quote,
                         expected)

    def test_base_differences_start_xrp_xlm_usdt_rates2(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 100,
                   'XLM': 100,
                   'USDT': 50, }
        rates = {'XRP/USDT': 0.1,
                 'XLM/USDT': 0.4,
                 'USDT/USDT': 1.0, }

        exchange = DummyExchange(targets.keys(), current, rates)
        portfolio = Portfolio.make_portfolio(targets, exchange)

        expected = {'XRP': 40,
                    'XLM': 0,
                    'USDT': -40, }
        self.assertEqual(portfolio.differences_quote,
                         expected)

        final_base = {}
        final_targets = {}
        total_base = 0
        for cur in current:
            symbol = "{}/{}".format(cur, portfolio.quote_currency)
            final_base[cur] = current[cur] * exchange.rates[symbol]['mid']
            final_base[cur] += portfolio.differences_quote[cur]
            total_base += final_base[cur]

        for cur in final_base:
            final_targets[cur] = (final_base[cur] / total_base) * 100

        self.assertEqual(targets, final_targets)

    def test_base_differences_start_xrp_xlm_usdt_rates3(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 40,
                   'XLM': 40,
                   'USDT': 20, }
        rates = {'XRP/USDT': 0.5,
                 'XLM/USDT': 0.5,
                 'USDT/USDT': 1.0, }

        exchange = DummyExchange(targets.keys(), current, rates)
        portfolio = Portfolio.make_portfolio(targets, exchange)

        expected = {'XRP': 7,
                    'XLM': 7,
                    'USDT': -14,
                    }
        self.assertEqual(portfolio.differences_quote,
                         expected)

        final_base = {}
        final_targets = {}
        total_base = 0
        for cur in current:
            symbol = "{}/{}".format(cur, portfolio.quote_currency)
            final_base[cur] = current[cur] * exchange.rates[symbol]['mid']
            final_base[cur] += portfolio.differences_quote[cur]
            total_base += final_base[cur]

        for cur in final_base:
            final_targets[cur] = (final_base[cur] / total_base) * 100

        self.assertEqual(targets, final_targets)


class test_SimpleBalancer(unittest.TestCase):

    def execute(self, targets, current, rates, fee=0.001, max_orders=5, mode='mid'):
        exchange = DummyExchange(targets.keys(), current, rates, fee)
        portfolio = Portfolio.make_portfolio(targets, exchange)

        balancer = SimpleBalancer()
        return balancer.balance(portfolio, exchange, max_orders=max_orders, mode=mode)

    def test_noop(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 450,
                   'XLM': 450,
                   'USDT': 100, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0, }

        res = self.execute(targets, current, rates)
        self.assertEqual(res['orders'], [])

    def test_start_all_usdt(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 0,
                   'XLM': 0,
                   'USDT': 1000, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates)

        expected = [Order('XLM/USDT', 'BUY', 400, 1.0),
                    Order('XRP/USDT', 'BUY', 500, 1.0), ]
        self.assertEqual(res['orders'], expected)

    def test_start_all_usdt_rates(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 0,
                   'XLM': 0,
                   'USDT': 1000, }
        rates = {'XRP/USDT': 0.5,
                 'XLM/USDT': 0.5, }

        res = self.execute(targets, current, rates)

        expected = [Order('XLM/USDT', 'BUY', 800, 0.5),
                    Order('XRP/USDT', 'BUY', 1000, 0.5), ]
        self.assertEqual(res['orders'], expected)

    def test_start_all_xrp(self):
        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 1000,
                   'XLM': 0,
                   'USDT': 0, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0, }
        res = self.execute(targets, current, rates)

        expected = [Order('XLM/XRP', 'BUY', 400, 1.0),
                    Order('XRP/USDT', 'SELL', 100, 1.0), ]
        self.assertEqual(res['orders'], expected)

    def test_start_xrp_xlm_usdt_rates(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 100,
                   'XLM': 100,
                   'USDT': 50, }
        rates = {'XRP/USDT': 0.1,
                 'XLM/USDT': 0.4, }
        res = self.execute(targets, current, rates)

        expected = [Order('XRP/USDT', 'BUY', 400, 0.1), ]
        self.assertEqual(res['orders'], expected)

    def test_mixed1(self):

        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 100,
                   'XLM': 100,
                   'USDT': 50, }
        rates = {'XRP/USDT': 0.1,
                 'XLM/USDT': 0.4, }
        res = self.execute(targets, current, rates)
        expected = [Order('XRP/USDT', 'BUY', 400, 0.1), ]
        self.assertEqual(res['orders'], expected)

    def test_mixed2a(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 400,
                   'XLM': 400,
                   'USDT': 200, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0, }
        res = self.execute(targets, current, rates)
        expected = [Order('XLM/USDT', 'BUY', 50, 1.0),
                    Order('XRP/USDT', 'BUY', 50, 1.0), ]
        self.assertEqual(res['orders'], expected)

    def test_mixed2b(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 800,
                   'XLM': 800,
                   'USDT': 200, }
        rates = {'XRP/USDT': 0.5,
                 'XLM/USDT': 0.5,
                 }
        res = self.execute(targets, current, rates)
        expected = [Order('XLM/USDT', 'BUY', 100, 0.5),
                    Order('XRP/USDT', 'BUY', 100, 0.5), ]
        self.assertEqual(res['orders'], expected)

    def test_mixed3a(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 500,
                   'XLM': 500,
                   'USDT': 0, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates)
        expected = [Order('XLM/USDT', 'SELL', 50, 1.0),
                    Order('XRP/USDT', 'SELL', 50, 1.0), ]
        self.assertEqual(res['orders'], expected)

    def test_mixed3b(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 1000,
                   'XLM': 1000,
                   'USDT': 0, }
        rates = {'XRP/USDT': 0.5,
                 'XLM/USDT': 0.5,
                 }

        res = self.execute(targets, current, rates)
        expected = [Order('XLM/USDT', 'SELL', 100, 0.5),
                    Order('XRP/USDT', 'SELL', 100, 0.5), ]
        self.assertEqual(res['orders'], expected)

    def test_real1a(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 6703.45,
                   'XLM': 0,
                   'USDT': 243.31, }
        base = "USDT"
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'USDT/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates)
        # Test the orders we get are correct
        expected = [Order('XLM/USDT', 'BUY', 165.12549537648613, 0.09084),
                    Order('XLM/XRP', 'BUY', 11691.753411492733, 0.283366),
                    Order('XRP/USDT', 'SELL', 34.42094463150016, 0.32076)]
        self.assertEqual(res['orders'], expected)

        # Test that the final amounts are in proportion to the targets
        base_amounts = {}
        for cur in res['proposed_portfolio'].currencies:
            pair = "{}/{}".format(cur, base)
            base_amounts[cur] = \
                res['proposed_portfolio'].balances[cur] * rates[pair]
        total_base = sum(base_amounts.values())
        for cur in targets:
            self.assertAlmostEqual(targets[cur],
                                   (base_amounts[cur] / total_base) * 100)

    def test_real2a(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3402.24,
                   'XLM': 6009.74,
                   'BTC': 0.14,
                   'ETH': 1.82,
                   'USDT': 270.82}
        base = "USDT"
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 'USDT/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates)
        # Test the orders we get are correct
        expected = [Order('BTC/USDT', 'SELL', 0.0037801180908891593, 3968.13),
                    Order('XLM/XRP', 'BUY', 6.551686481727605, 0.283366),
                    Order('XRP/BTC', 'SELL', 18.648636987155573, 8.102e-05),
                    Order('XRP/ETH', 'SELL', 13.236589350291911, 0.00217366),
                    Order('XRP/USDT', 'BUY', 39.15710063598936, 0.32076), ]
        self.assertEqual(res['orders'], expected)

        # Test that the final amounts are in proportion to the targets
        base_amounts = {}
        for cur in res['proposed_portfolio'].currencies:
            pair = "{}/{}".format(cur, base)
            base_amounts[cur] = \
                res['proposed_portfolio'].balances[cur] * rates[pair]
        total_base = sum(base_amounts.values())
        for cur in targets:
            self.assertAlmostEqual(targets[cur],
                                   (base_amounts[cur] / total_base) * 100)

    def test_real2a_cheap(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3402.24,
                   'XLM': 6009.74,
                   'BTC': 0.14,
                   'ETH': 1.82,
                   'USDT': 270.82}
        base = "USDT"
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 'USDT/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates, mode='cheap')
        # Test the orders we get are correct
        expected = [Order('XLM/XRP', 'BUY', 6.551686481727605, 0.283366),
                    Order('XRP/BTC', 'BUY', 28.11529866566897, 8.102e-05),
                    Order('XRP/ETH', 'SELL', 13.236589350292975, 0.00217366)] 
        self.assertEqual(res['orders'], expected)

    def test_real2a_cheaper(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3402.24,
                   'XLM': 6009.74,
                   'BTC': 0.14,
                   'ETH': 1.82,
                   'USDT': 270.82}
        base = "USDT"
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 'USDT/USDT': 1.0,
                 }

        res1 = self.execute(targets, current, rates, mode='mid')
        res2 = self.execute(targets, current, rates, mode='cheap')

        self.assertTrue(res2['total_fee'] < res1['total_fee'])

        
    def test_real2a_max_orders(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3352,
                   'XLM': 0,
                   'BTC': 0,
                   'ETH': 0,
                   'USDT': 243, }
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 }

        res = self.execute(targets, current, rates, max_orders=3)
        # Test the orders we get are correct
        expected = [Order('ETH/USDT', 'BUY', 0.7521902983559976, 147.81),
                    Order('XLM/XRP', 'BUY', 2902.218229854689, 0.283366),
                    Order('XRP/BTC', 'SELL', 821.9151515151515, 8.102e-05)]
        self.assertEqual(res['orders'], expected)

    def test_real2_nondirect(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3352,
                   'XLM': 0,
                   'BTC': 0,
                   'ETH': 0,
                   'USDT': 243, }
        base = "USDT"
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 'USDT/USDT': 1.0,
                 }

        res = self.execute(targets, current, rates)
        # Test the orders we get are correct
        expected = [Order('XLM/USDT', 'BUY', 1223.9239101717305, 0.09084),
                    Order('XRP/BTC', 'SELL', 821.9151515151515, 8.102e-05),
                    Order('XRP/ETH', 'SELL', 410.95757575757574, 0.00217366), ]

        self.assertEqual(res['orders'], expected)

        # Test that the final amounts are in proportion to the targets
        base_amounts = {}
        for cur in res['proposed_portfolio'].currencies:
            pair = "{}/{}".format(cur, base)
            base_amounts[cur] = \
                res['proposed_portfolio'].balances[cur] * rates[pair]
        total_base = sum(base_amounts.values())

        expected_targets = {'XRP': 51.56,
                            'XLM': 8.43,
                            'BTC': 20.00,
                            'ETH': 10.00,
                            'USDT': 10.00}

        for cur in targets:
            self.assertAlmostEqual(expected_targets[cur],
                                   (base_amounts[cur] / total_base) * 100,
                                   1)

    def test_badpair2(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3352,
                   'XLM': 0,
                   'BTC': 0,
                   'ETH': 0,
                   'USDT': 243, }
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 }

        with self.assertRaises(ValueError):
            self.execute(targets, current, rates)

    def test_badpair3(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 3352,
                   'XLM': 0,
                   'BTC': 0,
                   'ETH': 0,
                   'USDT': 243, }
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 }

        with self.assertRaises(ValueError):
            self.execute(targets, current, rates)

    def test_zero_balance(self):

        targets = {'XRP': 40,
                   'XLM': 20,
                   'BTC': 20,
                   'ETH': 10,
                   'USDT': 10, }
        current = {'XRP': 0,
                   'XLM': 0,
                   'BTC': 0,
                   'ETH': 0,
                   'USDT': 0, }
        rates = {'XRP/USDT': 0.32076,
                 'XLM/USDT': 0.09084,
                 'XLM/XRP': 0.283366,
                 'XRP/BTC': 0.00008102,
                 'XRP/ETH': 0.00217366,
                 'BTC/USDT': 3968.13,
                 'ETH/USDT': 147.81,
                 }

        res = self.execute(targets, current, rates)
        # Test the orders we get are correct
        expected = []
        self.assertEqual(res['orders'], expected)

    def test_fees1(self):
        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 0,
                   'XLM': 0,
                   'USDT': 1000, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 }
        res = self.execute(targets, current, rates)
        self.assertAlmostEqual(res['total_fee'], 0.9)

    def test_fees2(self):
        targets = {'XRP': 50,
                   'XLM': 40,
                   'USDT': 10, }
        current = {'XRP': 0,
                   'XLM': 0,
                   'USDT': 1000, }
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 }
        res = self.execute(targets, current, rates, 0.005)
        self.assertEqual(res['total_fee'], 4.5)


class test_Executor(unittest.TestCase):

    def create_executor(self, targets, current, rates, fee=0.001):
        exchange = DummyExchange(targets.keys(), current, rates, fee)
        portfolio = Portfolio.make_portfolio(targets, exchange)
        balancer = SimpleBalancer()
        executor = Executor(portfolio, exchange, balancer)
        return executor

    def test_threshold_inbalance(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        executor = self.create_executor(targets, current, rates)
        res = executor.run()

        # Test the orders we get are correct
        expected = []
        self.assertEqual(res['orders'], expected)

    def test_threshold_under(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 46,
                   'XLM': 45,
                   'USDT': 9}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        executor = self.create_executor(targets, current, rates)
        res = executor.run()

        # Test the orders we get are correct
        expected = []
        self.assertEqual(res['orders'], expected)

    def test_threshold_under_force(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 460,
                   'XLM': 450,
                   'USDT': 90}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        executor = self.create_executor(targets, current, rates)
        res = executor.run(force=True)

        # Test the orders we get are correct
        expected = [Order('XRP/USDT', 'SELL', 10, 1.0)]
        self.assertEqual(res['orders'], expected)

    def test_threshold_over(self):

        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 400,
                   'XLM': 400,
                   'USDT': 200}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        executor = self.create_executor(targets, current, rates)
        res = executor.run(force=False)

        # TEST the orders we get are correct
        expected = [Order('XLM/USDT', 'BUY', 50, 1.0),
                    Order('XRP/USDT', 'BUY', 50, 1.0), ]
        self.assertEqual(res['orders'], expected)

    def test_run_notrade(self):
        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 400,
                   'XLM': 400,
                   'USDT': 200}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        exchange = DummyExchange(targets.keys(), current, rates, 0.001)
        portfolio = Portfolio.make_portfolio(targets, exchange)
        balancer = SimpleBalancer()
        executor = Executor(portfolio, exchange, balancer)

        res = executor.run(force=True, trade=False)

        # Test the orders we get are correct
        expected = [Order('XLM/USDT', 'BUY', 50.0, 1.0),
                    Order('XRP/USDT', 'BUY', 50.0, 1.0), ]
        self.assertEqual(res['orders'], expected)

        self.assertEqual(exchange.balances['XRP'], 400)
        self.assertEqual(exchange.balances['XLM'], 400)
        self.assertEqual(exchange.balances['USDT'], 200)

    def test_run_trade(self):
        targets = {'XRP': 45,
                   'XLM': 45,
                   'USDT': 10, }
        current = {'XRP': 400,
                   'XLM': 400,
                   'USDT': 200}
        rates = {'XRP/USDT': 1.0,
                 'XLM/USDT': 1.0,
                 'XLM/XRP': 1.0,
                 }

        exchange = DummyExchange(targets.keys(), current, rates, 0.001)
        portfolio = Portfolio.make_portfolio(targets, exchange)
        balancer = SimpleBalancer()
        executor = Executor(portfolio, exchange, balancer)

        res = executor.run(force=True, trade=True)

        # Test the orders we get are correct
        expected = [Order('XLM/USDT', 'BUY', 50, 1.0),
                    Order('XRP/USDT', 'BUY', 50, 1.0), ]
        self.assertEqual(res['orders'], expected)

        self.assertEqual(exchange.balances['XRP'], 449.95)
        self.assertEqual(exchange.balances['XLM'], 449.95)
        self.assertEqual(exchange.balances['USDT'], 100)


class test_DummyExchange(unittest.TestCase):
    def setUp(self):
        balances = {'XRP': 100.0,
                    'BTC': 200.0,
                    'USDT': 300.0}
        rates = {'XRP/USDT': 0.33,
                 'BTC/USDT': 3500.0}
        self.exchange = DummyExchange(balances.keys(), balances, rates)

    def test_balances(self):
        self.assertEqual(self.exchange.balances['XRP'], 100)
        self.assertEqual(self.exchange.balances['BTC'], 200)
        self.assertEqual(self.exchange.balances['USDT'], 300)

    def test_pairs(self):
        expected = ['XRP/XRP', 'XRP/BTC', 'XRP/USDT',
                    'BTC/XRP', 'BTC/BTC', 'BTC/USDT',
                    'USDT/XRP', 'USDT/BTC', 'USDT/USDT']
        self.assertEqual(self.exchange.pairs, expected)

    def test_execute_buy(self):
        order = Order('XRP/USDT', 'BUY', 10, 0.32)
        self.exchange.execute_order(order)
        self.assertEqual(self.exchange.balances['XRP'], 109.99)
        self.assertEqual(self.exchange.balances['BTC'], 200.0)
        self.assertEqual(self.exchange.balances['USDT'], 296.8)

    def test_execute_sell(self):
        order = Order('BTC/USDT', 'SELL', 0.01, 3500)
        self.exchange.execute_order(order)
        self.assertEqual(self.exchange.balances['XRP'], 100.00)
        self.assertEqual(self.exchange.balances['BTC'], 199.99)
        self.assertEqual(self.exchange.balances['USDT'], 334.965)

    def test_execute_buy_toomuch1(self):
        order = Order('XRP/USDT', 'BUY', 1000, 0.32)
        with self.assertRaises(ValueError):
            self.exchange.execute_order(order)

    def test_execute_sell_toomuch2(self):
        order = Order('BTC/USDT', 'SELL', 1000, 3500)
        with self.assertRaises(ValueError):
            self.exchange.execute_order(order)

    def test_preprocess_order_true(self):
        order = Order('XRP/USDT', 'BUY', 50, 0.32)
        self.assertEqual(self.exchange.preprocess_order(order), order)

    def test_preprocess_order_false(self):
        order = Order('XRP/USDT', 'BUY', 0.1, 0.32)
        self.assertIsNone(self.exchange.preprocess_order(order))

    def test_preprocess_order_bad(self):
        order = Order('ZEC/USDT', 'BUY', 10, 0.32)
        self.assertIsNone(self.exchange.preprocess_order(order))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from crypto_reflex.backtest_exchange import BacktestExchange
from crypto_reflex.simple_balancer import SimpleBalancer
from crypto_reflex.portfolio import Portfolio

if __name__ == '__main__':
    Xbalances = {'XRP':3269.878282,
                'BTC': 0.13551801,
                'ETH': 3.78736025,
                'BNB': 25.08296788,
                'USDT': 251.05799855}

    balances = {'XRP': 50000.0,
                'BTC': 0.0,
                'ETH': 0.0,
                'USD': 0.0}

    balances = {'XRP': 0.0,
                'BTC': 0.0,
                'ETH': 0.0,
                'USD': 10000.0}

    Xtargets = {'XRP': 40,
               'BTC': 20,
               'ETH': 20,
               'BNB': 10,
               'USDT': 10, }

    Xtargets = {'XRP': 20,
               'BTC': 20,
               'ETH': 20,
               'BNB': 20,
               'USDT': 20, }

    Xtargets = {'XRP': 40,
               'BTC': 20,
               'ETH': 20,
               'USD': 20, }

    targets = {'XRP': 80,
               'USD': 20, }
    
    for t in range(10,100,10):
        threshold = t / 10.0
        exchange = BacktestExchange('/Development/crypto_reflex/data/*.json', balances.copy())
        portfolio = Portfolio.make_portfolio(targets, exchange, threshold, quote_currency="USD")
        balancer = SimpleBalancer()
        num_trades = 0

        while portfolio.needs_balancing:
            res = balancer.balance(portfolio,
                                   exchange,
                                   accuracy=True,
                                   max_orders=4)
            for order in res['orders']:
                try:
                    r = exchange.execute_order(order)
                    num_trades += 1
                except ValueError:
                    pass

            portfolio.sync_balances()
            
        initial_portfolio = portfolio.copy()
        
        while True:

            if portfolio.needs_balancing:
                res = balancer.balance(portfolio,
                                       exchange,
                                       accuracy=True,
                                       max_orders=2)
                for order in res['orders']:
                    try:
                        r = exchange.execute_order(order)
                        num_trades += 1
                    except ValueError:
                        pass

                portfolio.sync_balances()
                
            try:
                exchange.tick()
            except StopIteration:
                break
            portfolio.sync_rates()

        print("Threshold:", threshold)
        print("Initial value:", initial_portfolio.valuation_quote)
        print("Final value: ", portfolio.valuation_quote)

        portfolio.balances = initial_portfolio.balances

        print("B&H value: ", portfolio.valuation_quote)
        print("Number of trades: ", num_trades)
        print()
    



import json
class Order():
    def __init__(self, pair, direction, amount, price):
        if direction.upper() not in ['BUY', 'SELL']:
            raise ValueError("{} is not a valid direction".format(direction))
        self.pair = pair
        self.direction = direction
        self.amount = float(amount)
        self.price = float(price)
        self.type_ = None

    def __str__(self):
        return f"{self.direction} {self.amount} {self.pair} @ {self.price}"

    def __repr__(self):
        return f"Order('{self.pair}', '{self.direction}', {self.amount}, {self.price})"

    def __eq__(self, other):
        return self.pair == other.pair and \
            self.direction == other.direction and \
            self.amount == other.amount and \
            self.price == other.price

    def __lt__(self, other):
        return (self.pair, self.direction, self.amount, self.price) < \
            (other.pair, other.direction, other.amount, other.price)

    def __hash__(self):
        return hash((self.pair, self.direction, self.amount, self.price))

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, 
            sort_keys=True, indent=4)



import argparse
import configparser
import logging
import sys
import json

from crypto_reflex.simple_balancer import SimpleBalancer
from crypto_reflex.ccxt_exchange import CCXTExchange, exchanges
from crypto_reflex.executor import Executor
from crypto_reflex.portfolio import Portfolio

logger = logging.getLogger(__name__)


def main(args=None):
    config = configparser.ConfigParser()
    config.read('config.ini')

    def exchange_choices():
        return set(config.sections()) & set(exchanges)

    parser = argparse.ArgumentParser(
        description='Balance holdings on an exchange.')
    parser.add_argument('--trade', action="store_true",
                        help='Actually place orders')
    parser.add_argument('-j','--json', action='store_true',
                        help='returns the data as a json')
    parser.add_argument('--force', action="store_true",
                        help='Force rebalance')
    parser.add_argument('--max_orders', default=5,
                        help='Maximum number of orders to perform in '
                             'rebalance')
    parser.add_argument('--valuebase', default='USDT',
                        help='Currency to value portfolio in')
    parser.add_argument('--cancel', action="store_true",
                        help='Cancel open orders first')
    parser.add_argument('--mode', choices=['mid', 'passive', 'cheap'],
                        default='mid',
                        help='Mode to place orders')
    parser.add_argument('exchange', choices=exchange_choices())
    args = parser.parse_args()

    config = config[args.exchange]

    try:
        targets = [x.split() for x in config['targets'].split('\n')]
        targets = dict([[a, float(b)] for (a, b) in targets])
    except ValueError:
        logger.error("Targets format invalid")
        sys.exit(1)

    total_target = sum(targets.values())
    if total_target != 100:
        logger.error("Total target needs to equal 100, it is {}"
                     .format(total_target))
        sys.exit(1)

    valuebase = config.get('valuebase') or args.valuebase

    exchange = CCXTExchange(args.exchange,
                            targets.keys(),
                            config['api_key'],
                            config['api_secret'])
    if args.json == False:
        print("Connected to exchange: {}".format(exchange.name))
        print()

    if args.cancel:
        print("Cancelling open orders...")
        for order in exchange.cancel_orders():
            print("Cancelled order:", order['symbol'], order['id'])
        print()

    threshold = float(config['threshold'])
    max_orders = int(args.max_orders)

    portfolio = Portfolio.make_portfolio(targets, exchange, threshold, valuebase)
    curr_portafolio = portfolio
    if args.json == False:
        print("Current Portfolio:")
    for cur in portfolio.balances:
        bal = portfolio.balances[cur]
        pct = portfolio.balances_pct[cur]
        tgt = targets[cur]
        if args.json == False:
            print("  {:<6s} {:<8.2f} ({:>5.2f} / {:>5.2f}%)"
                .format(cur, bal, pct, tgt))

            print()
            print("  Total value: {:.16f} {}".format(portfolio.valuation_quote,
                                            portfolio.quote_currency))
    balancer = SimpleBalancer()
    executor = Executor(portfolio, exchange, balancer)
    res = executor.run(force=args.force,
                       trade=args.trade,
                       max_orders=max_orders,
                       mode=args.mode)
    if args.json == False:
        print("  Balance RMS error: {:.2g} / {:.2g}".format(
            res['initial_portfolio'].balance_rms_error,
            threshold))

        print("  Balance Max error: {:.2g} / {:.2g}".format(
            res['initial_portfolio'].balance_max_error,
            threshold))
    
        print()
    if not portfolio.needs_balancing and not args.force:
        print("No balancing needed")
        sys.exit(0)
    if args.json == False:
        print("Balancing needed{}:".format(" [FORCED]" if args.force else ""))
        print()
        print("Proposed Portfolio:")
    portfolio = res['proposed_portfolio']

    if not portfolio:
        print("Could not calculate a better portfolio")
        sys.exit(0)

    for cur in portfolio.balances:
        bal = portfolio.balances[cur]
        pct = portfolio.balances_pct[cur]
        tgt = targets[cur]
        if args.json == False:
            print("  {:<6s} {:<8.16f} ({:>5.16f} / {:>5.16f}%)"
                .format(cur, bal, pct, tgt))

            print()
            print("  Total value: {:.2f} {}".format(portfolio.valuation_quote,
                                            portfolio.quote_currency))
            print("  Balance RMS error: {:.2g} / {:.2g}".format(
                res['proposed_portfolio'].balance_rms_error,
                threshold))

            print("  Balance Max error: {:.2g} / {:.2g}".format(
                res['proposed_portfolio'].balance_max_error,
                threshold))

    total_fee = '%s' % float('%.4g' % res['total_fee'])
    if args.json == False:
        print("  Total fees to re-balance: {} {}"
            .format(total_fee,
                    portfolio.quote_currency))

        print()
        print("Orders:")
        if args.trade:
            for order in res['success']:
                print("  Submitted: {}".format(order))

            for order in res['errors']:
                print("  Failed: {}".format(order))
        else:
            for order in res['orders']:
                print("  " + str(order))


    if args.json:
        return json.dumps({"portfolio_value": portfolio.valuation_quote, "currency": portfolio.quote_currency, "cost": total_fee})

if __name__ == '__main__':
    main()



from crypto_reflex.order import Order
from itertools import product

class Attempt():
    def __init__(self, portfolio, orders=None, total_fee=0.0, depth=0,
                 pairs_processed=None):
        self.portfolio = portfolio
        self.orders = orders or []
        self.total_fee = total_fee
        self.depth = depth
        self.pairs_processed = pairs_processed or set()


class SimpleBalancer():

    def permute_differences(self, differences_quote):
        differences = differences_quote.items()
        positives = [x for x in differences if x[1] > 0]
        negatives = [x for x in differences if x[1] < 0]
        res = product(positives, negatives)
        return res

    def balance(self, initial_portfolio, exchange, max_orders=5, mode='mid'):
        rates = exchange.rates
        quote_currency = initial_portfolio.quote_currency

        # Add in the identify rate just so we don't have to special
        # case it later
        rates["{}/{}".format(quote_currency, quote_currency)] = {'mid': 1.0,
                                                                 'high': 1.0,
                                                                 'low': 1.0, }

        todo = [Attempt(initial_portfolio)]
        attempts = []

        while todo:
            attempt = todo.pop()
            diffs = self.permute_differences(
                attempt.portfolio.differences_quote)

            if not diffs or attempt.depth >= max_orders:
                continue

            for pos, neg in diffs:
                p_cur, p_amount = pos
                n_cur, n_amount = neg

                trade_direction = None

                # try and see if we have the pair we need
                # and if so, buy it
                pair = "{}/{}".format(p_cur, n_cur)
                if pair in rates:
                    trade_direction = "BUY"
                    trade_pair = pair

                # if previous failed then reverse paid and try
                # sell instead
                pair = "{}/{}".format(n_cur, p_cur)
                if pair in rates:
                    trade_direction = "SELL"
                    trade_pair = pair

                if not trade_direction:
                    continue

                if trade_pair in attempt.pairs_processed:
                    continue

                # Calculate the min order for this pair
                min_trade_amount_quote = \
                    exchange.limits[trade_pair]['cost']['min']

                # Work out the pair to get to the quote currency
                to_sell_pair_quote = "{}/{}".format(n_cur, quote_currency)
                to_buy_pair_quote = "{}/{}".format(p_cur, quote_currency)

                amounts = [p_amount, -n_amount,]
                if mode == 'mid':
                    amounts.extend([min_trade_amount_quote, min_trade_amount_quote * 1.5])

                for trade_amount_quote in amounts:

                    # Work out how much of the currency to buy/sell
                    to_sell_amount_cur = \
                        trade_amount_quote / rates[to_sell_pair_quote]['mid']
                    to_buy_amount_cur = \
                        trade_amount_quote / rates[to_buy_pair_quote]['mid']

                    if trade_direction == "BUY":
                        trade_amount = to_buy_amount_cur

                    if trade_direction == "SELL":
                        trade_amount = to_sell_amount_cur

                    # We got a direction, so we know we can either
                    # buy or sell this pair
                    if mode == 'passive':
                        if trade_direction == 'BUY':
                            trade_rate = rates[trade_pair]['low']
                        if trade_direction == 'SELL':
                            trade_rate = rates[trade_pair]['high']                         
                    else:
                        trade_rate = rates[trade_pair]['mid']

                    order = Order(trade_pair, trade_direction,
                                  trade_amount, trade_rate)

                    order = exchange.preprocess_order(order)
                    if not order:
                        continue

                    # Adjust the amounts of each currency we hold
                    new_portfolio = attempt.portfolio.copy()

                    new_portfolio.balances[p_cur] \
                        += to_buy_amount_cur
                    new_portfolio.balances[n_cur] \
                        -= to_sell_amount_cur

                    if new_portfolio.balances[n_cur] < 0:
                        # gone negative so not valid result
                        break

                    fee = trade_amount_quote * exchange.fee
                    new_attempt = Attempt(new_portfolio,
                                          sorted(attempt.orders + [order]),
                                          attempt.total_fee + fee,
                                          attempt.depth + 1,
                                          attempt.pairs_processed | set(
                                              [trade_pair]
                                          ))
                    todo.append(new_attempt)

                    if new_attempt.portfolio.balance_rms_error < \
                       initial_portfolio.balance_rms_error:
                        attempts.append(new_attempt)

        if not attempts:
            return {'orders': [],
                    'total_fee': 0.0,
                    'initial_portfolio': initial_portfolio,
                    'proposed_portfolio': None}

        sort_key = lambda x: (x.portfolio.balance_rms_error,
                              x.total_fee,
                              len(x.orders),
                              x.orders)
        decorated_attempts = [(sort_key(x), x) for x in attempts]
        decorated_attempts.sort(key=lambda x: x[0])
        best_attempt = decorated_attempts[0][1]

        return {'orders': sorted(best_attempt.orders),
                'total_fee': best_attempt.total_fee,
                'initial_portfolio': initial_portfolio,
                'proposed_portfolio': best_attempt.portfolio}



import math


class Portfolio():

    @classmethod
    def make_portfolio(cls, targets, exchange,
                       threshold=1.0, quote_currency="USDT"):
        p = cls(targets, exchange, threshold, quote_currency)
        p.sync_balances()
        p.sync_rates()
        return p

    def __init__(self, targets, exchange, threshold=1.0,
                 quote_currency="USDT"):
        self.targets = targets
        self.threshold = threshold
        self.exchange = exchange
        self.quote_currency = quote_currency
        self.balances = {}
        self.rates = {}

    def copy(self):
        p = Portfolio(self.targets,
                      self.exchange,
                      self.threshold,
                      self.quote_currency)
        p.balances = self.balances.copy()
        p.rates = self.rates.copy()
        return p

    def sync_balances(self):
        self.balances = self.exchange.balances.copy()

    def sync_rates(self):
        self.rates = self.exchange.rates.copy()

    @property
    def currencies(self):
        return self.targets.keys()

    @property
    def balances_quote(self):
        _balances_quote = {}
        qc = self.quote_currency
        for cur in self.currencies:
            amount = self.balances[cur]
            if cur == self.quote_currency:
                _balances_quote[cur] = amount
            else:
                pair = f"{cur}/{qc}"
                try:
                    _balances_quote[cur] = amount * self.rates[pair]['mid']
                except KeyError:
                    raise ValueError("Invalid pair: {}".format(pair))

        return _balances_quote

    @property
    def valuation_quote(self):
        return sum(self.balances_quote.values())

    @property
    def needs_balancing(self):
        return self.balance_max_error > self.threshold

    @property
    def balances_pct(self):
        # first convert the amounts into their base value
        _balances_quote = self.balances_quote
        _total = self.valuation_quote

        if not _total:
            return {cur: 0 for cur in self.currencies}

        return {cur: (_balances_quote[cur] / _total) * 100.0
                for cur in self.currencies}

    @property
    def balance_errors_pct(self):
        _balances_quote = self.balances_quote
        _total = sum(_balances_quote.values())

        if not _total:
            return []

        def calc_diff(cur):
            return _total * (self.targets[cur] / 100.0) \
                - _balances_quote[cur]

        pcts = [(calc_diff(cur) / _total) * 100.0
                for cur in self.currencies]
        return pcts

    @property
    def balance_rms_error(self):
        pcts = self.balance_errors_pct
        num = len(pcts)
        if not num:
            return 0.0
        return math.sqrt(sum([x**2 for x in pcts]) / num)

    @property
    def balance_max_error(self):
        pcts = self.balance_errors_pct
        pcts = [abs(x) for x in pcts]
        return max(pcts)

    @property
    def differences_quote(self):
        # first convert the amounts into their base value
        _balances_quote = self.balances_quote
        _total = self.valuation_quote

        def calc_diff(cur):
            return _total * (self.targets[cur] / 100.0) \
                - _balances_quote[cur]

        return {cur: calc_diff(cur) for cur in self.currencies}



import ccxt, os

from functools import lru_cache

exchanges = ccxt.exchanges


class CCXTExchange():

    def __init__(self, name, currencies, api_key, api_secret):
        self.name = name
        self.currencies = currencies
        self.exch = getattr(ccxt, name)({'nonce': ccxt.Exchange.milliseconds})
        if os.getenv("PROXY_URL") is not None:
            self.exch.proxies= {
                "http"  : os.environ.get('PROXY_URL', ''),
                "https" : os.environ.get('PROXY_URL', '')
            }
        self.exch.apiKey = api_key
        self.exch.secret = api_secret
        self.exch.load_markets()

    @property
    @lru_cache(maxsize=None)
    def balances(self):
        bals = self.exch.fetch_balance()['total']
        return {k: bals[k] for k in self.currencies}

    @property
    @lru_cache(maxsize=None)
    def pairs(self):
        _pairs = []
        for i in self.currencies:
            for j in self.currencies:
                pair = "{}/{}".format(i, j)
                if pair in self.exch.markets and self.exch.markets[pair]['active']:
                    _pairs.append(pair)
        return _pairs

    @property
    @lru_cache(maxsize=None)
    def rates(self):
        _rates = {}
        if self.exch.has['fetchTickers']:
            tickers = self.exch.fetchTickers()
        else:
            tickers = {}

        for pair in self.pairs:
            if tickers:
                high = tickers[pair]['ask']
                low = tickers[pair]['bid']
            else:
                orderbook = self.exch.fetchOrderBook(pair)
                high = orderbook['asks'][0][0]
                low = orderbook['bids'][0][0]
            mid = (high + low) / 2.0
            _rates[pair] = {'mid': mid,
                            'high': high,
                            'low': low, }

        return _rates

    @property
    @lru_cache(maxsize=None)
    def limits(self):
        return {pair: self.exch.markets[pair]['limits']
                for pair in self.pairs}

    @property
    @lru_cache(maxsize=None)
    def fee(self):
        return self.exch.fees['trading']['maker']

    def preprocess_order(self, order):
        try:
            limits = self.limits[order.pair]
        except KeyError:
            return None

        order.amount = float(
            self.exch.amount_to_precision(
                order.pair, order.amount))
        order.price = float(
            self.exch.price_to_precision(
                order.pair, order.price))

        if order.price == 0 or order.amount == 0:
            return None

        if order.amount < limits['amount']['min'] \
           or order.amount * order.price < limits['cost']['min']:
            return None
        order.type_ = 'LIMIT'
        return order

    def execute_order(self, order):
        if not order.type_:
            raise ValueError("Order needs preprocessing first")
        return self.exch.create_order(order.pair,
                                      order.type_,
                                      order.direction,
                                      order.amount,
                                      order.price)

    def cancel_orders(self):
        cancelled_orders = []
        for pair in self.pairs:
            open_orders = self.exch.fetch_open_orders(symbol=pair)
            for order in open_orders:
                self.exch.cancel_order(order['id'], order['symbol'])
                cancelled_orders.append(order)
        return cancelled_orders



import os
from setuptools import setup

base_path = os.path.dirname(__file__)
with open(os.path.join(base_path, "README.md"), encoding='utf-8') as readme_file:
    readme = readme_file.read()

setup(name='crypto_reflex',
      version='1.0.0',
      packages=['crypto_reflex'],
      entry_points={
          'console_scripts': [
              'crypto_reflex = crypto_reflex.main:main'
          ]
      },
      license='MIT',
      description = 'A tool to automatically balance cryptocurrency portfolios',
      long_description=readme,
      long_description_content_type='text/markdown',
      author = 'Draczer01',
      author_email = 'draczer01@gmail.com',
      url = 'https://github.com/draczer01/Crypto_Reflex',
      keywords = ['cryptocurrency', 'portfolio', 'xrp', 'ethereum', 'bitcoin', 'btc', 'eth','balancer','rebalancer'],
      install_requires=[
          'ccxt',
      ],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Financial and Insurance Industry',
          'Topic :: Office/Business :: Financial :: Investment',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.9',
      ],
      )



import glob
import json
import pandas as pd

from crypto_reflex.dummy_exchange import DummyExchange


LIMITS = {'BNB/BTC': {'amount': {'max': 90000000.0, 'min': 0.01},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'BNB/ETH': {'amount': {'max': 90000000.0, 'min': 0.01},
                      'cost': {'max': None, 'min': 0.01},
                      'price': {'max': None, 'min': None}},
          'BNB/USD': {'amount': {'max': 10000000.0, 'min': 0.01},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'BTC/USD': {'amount': {'max': 10000000.0, 'min': 1e-06},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'ETH/BTC': {'amount': {'max': 100000.0, 'min': 0.001},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'ETH/USD': {'amount': {'max': 10000000.0, 'min': 1e-05},
                       'cost': {'max': None, 'min': 10.0},
                       'price': {'max': None, 'min': None}},
          'XRP/BNB': {'amount': {'max': 90000000.0, 'min': 0.1},
                      'cost': {'max': None, 'min': 1.0},
                      'price': {'max': None, 'min': None}},
          'XRP/BTC': {'amount': {'max': 90000000.0, 'min': 1.0},
                      'cost': {'max': None, 'min': 0.001},
                      'price': {'max': None, 'min': None}},
          'XRP/ETH': {'amount': {'max': 90000000.0, 'min': 1.0},
                      'cost': {'max': None, 'min': 0.01},
                      'price': {'max': None, 'min': None}},
          'XRP/USD': {'amount': {'max': 90000000.0, 'min': 0.1},
                       'cost': {'max': None, 'min': 1.0},
                       'price': {'max': None, 'min': None}},
          'XLM/USD': {'amount': {'max': 90000000.0, 'min': 0.1},
                       'cost': {'max': None, 'min': 1.0},
                       'price': {'max': None, 'min': None}},
          'XLM/XRP': {'amount': {'max': 90000000.0, 'min': 0.1},
                      'cost': {'max': None, 'min': 1.0},
                      'price': {'max': None, 'min': None}}}


class BacktestExchange(DummyExchange):
    
    def __init__(self, filenames, balances, fee=0.001):
        self.name = 'BacktestExchange'
        self._currencies = balances.keys()

        final_df = pd.DataFrame()
        for path in glob.glob(filenames):
            filename = path.split('/')[-1]
            pair = filename.split('.')[0]
            pair = pair.replace('-','/')
            self.pairs.append(pair)
            data = json.load(open(path, 'r'))#['Data']
            df = pd.DataFrame(data)
            df.set_index(pd.to_datetime(df['time'], unit='s'), inplace=True)
            df = df[~df.index.duplicated()]
            final_df[pair] = df['close']
#            print('loaded', pair)

        final_df.fillna(method='ffill', inplace=True)
#        self._iter = final_df['2018-01-08':].iterrows()
#        self._iter = final_df['2017-09-11':].iterrows()
#        self._iter = final_df[:'2018-12-30'].iterrows()
#        self._iter = final_df['2018-09-11':'2018-12-30'].iterrows()
        self._iter = final_df.iterrows()
        self._rates = {}
        
        self._balances = balances
        self._fee = fee

        self.tick()

    def tick(self):
        self._rates = dict(next(self._iter)[1])

    @property
    def limits(self):
        return LIMITS
